# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from .auth import (
    get_app_access_token,
    get_app_access_token_internal,
    get_app_access_token_store,
    get_tenant_access_token,
    get_tenant_access_token_internal,
    get_tenant_access_token_store,
)
from .encrypt import AESCipher
from .exceptions import FeishuException
from .request import delete, get, patch, post, put

__all__ = [
    "post",
    "get",
    "put",
    "patch",
    "delete",
    "AESCipher",
    "FeishuException",
    "get_tenant_access_token",
    "get_tenant_access_token_internal",
    "get_tenant_access_token_store",
    "get_app_access_token",
    "get_app_access_token_internal",
    "get_app_access_token_store",
]



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=unused-argument, too-many-arguments

from __future__ import annotations

import time
from typing import Callable, Dict

import requests
from chanfig import NestedDict

from feishu.api import FeishuException, variables
from feishu.api.auth import get_tenant_access_token as _get_tenant_access_token

token_cache = {"tenant_access_token": None, "expire": 0, "timestamp": 0}


def get_tenant_access_token(app_id: str, app_secret: str, timeout: int = 120) -> str:
    current_time = int(time.time())
    token: str = token_cache["tenant_access_token"]  # type: ignore[assignment]

    if token is None or token_cache["expire"] < current_time:  # type: ignore[operator]
        response = _get_tenant_access_token(app_id, app_secret, timeout)
        token = token_cache["tenant_access_token"] = response["tenant_access_token"]
        token_cache["expire"] = current_time + response["expire"] - variables.ACCESS_TOKEN_REFRESH_OFFSET
        token_cache["timestamp"] = current_time
    return token


def request(
    method: str | Callable,
    dest: str,
    data: Dict | None = None,
    params: Dict | None = None,
    headers: Dict | None = None,
    timeout: int = 120,
    token: str | None = None,
    app_id: str | None = None,
    app_secret: str | None = None,
    max_retries: int | None = None,
    backoff_factor: float | None = None,
    retry_codes: tuple[int, ...] = (500, 502, 503, 504),
    **kwargs,
) -> NestedDict:
    r"""
    向指定目的地发送请求。

    Args:
        method: 请求方法。可以是字符串（'GET'、'POST'、'PUT'、'PATCH'、'DELETE'）或者函数。
        dest: 目的地统一资源定位符。这个地址应该是相对于飞书 API 的。
        data: 请求数据。默认为 `None`。
        params: 请求参数。默认为 `None`。
        headers: 请求头。默认为 `{}`。
        timeout: 超时时间。默认为 120 秒。
        token: 访问凭证。如果为 `None`，则会自动获取。
            如果 token 为 `None`，则 app_id 和 app_secret 必须不为 `None`。
        app_id: 应用唯一标识（以`cli_`开头）。默认为 `None`。
        app_secret: 应用秘钥。默认为 `None`。
        max_retries: 最大重试次数。默认为 `variables.MAX_RETRIES`。
        backoff_factor: 重试间隔因子。默认为 `variables.BACKOFF_FACTOR`。
        retry_codes: 需要重试的状态码。默认为 `(500, 502, 503, 504)`
        kwargs: 其他参数。

    Returns:
        请求结果

    Raises:
        ValueError: 如果 token, app_id, 和 app_secret 都为 `None`。
        FeishuException: 飞书 API 返回的错误。
    """
    if max_retries is None:
        max_retries = variables.MAX_RETRIES
    if backoff_factor is None:
        backoff_factor = variables.BACKOFF_FACTOR
    if dest.startswith("/"):
        dest = dest[1:]
    url = variables.BASE_URL + dest
    if headers is None:
        headers = {}
    if token is None:
        if app_id is None or app_secret is None:
            raise ValueError("token, app_id, and app_secret cannot all be None.")
        token = get_tenant_access_token(app_id, app_secret, timeout)
    headers["Authorization"] = f"Bearer {token}"
    if data is not None:
        headers["Content-Type"] = "application/json"

    attempt = 0
    while attempt <= max_retries:
        if callable(method):
            response = method(url, headers=headers, json=data, params=params, timeout=timeout, **kwargs)
        else:
            response = requests.request(
                method, url, headers=headers, json=data, params=params, timeout=timeout, **kwargs
            )
        ret = response.json()
        if ret.get("code") == 0:
            return NestedDict(ret)
        if response.status_code in retry_codes:
            time.sleep(backoff_factor**attempt)
            attempt += 1
        raise FeishuException(ret.get("code"), ret.get("msg"))

    raise FeishuException(ret.get("code"), f"Request failed after {max_retries} retries.")


def post(
    dest: str,
    data: Dict | None = None,
    params: Dict | None = None,
    headers: Dict | None = None,
    timeout: int = 120,
    token: str | None = None,
    app_id: str | None = None,
    app_secret: str | None = None,
    **kwargs,
) -> NestedDict:
    r"""
    向指定目的地发送 POST 请求。

    Args:
        dest: 目的地统一资源定位符。这个地址应该是相对于飞书 API 的。
        headers: 请求头。默认为 `{}`。
        params: 请求参数。默认为 `{}`。
        timeout: 超时时间。默认为 120 秒。
        token: 访问凭证。如果为 `None`，则会自动获取。
            如果 token 为 `None`，则 app_id 和 app_secret 必须不为 `None`。
        app_id: 应用唯一标识（以`cli_`开头）。默认为 `None`。
        app_secret: 应用秘钥。默认为 `None`。

    Returns:
        请求结果

    Raises:
        ValueError: 如果 token, app_id, 和 app_secret 都为 `None`。

    Examples:
        >>> data = {
        ...     "receive_id_type": "open_id",
        ...     "receive_id": "ou_7d8a6e6df7621556ce0d21922b676706ccs",
        ...     "msg_type": "text",
        ...     "content": "{\"text\":\"test content\"}",
        ...     "uuid": "a0d69e20-1dd1-458b-k525-dfeca4015204"
        ... }
        >>> post('im/v1/messages', data=data, app_id='cli_slkdjalasdkjasd', app_secret='dskLLdkasdjlasdKK')  # doctest:+SKIP
        NestedDict(
          ('code'): 0
          ('msg'): 'success'
          ('data'): NestedDict(
            ('message_id'): 'om_dc13264520392913993dd051dba21dcf'
            ('root_id'): 'om_40eb06e7b84dc71c03e009ad3c754195'
            ('parent_id'): 'om_d4be107c616aed9c1da8ed8068570a9f'
            ('thread_id'): 'omt_d4be107c616a'
            ('msg_type'): 'card'
            ('create_time'): '1635675360'
            ('update_time'): '1635675360'
            ('deleted'): False
            ('updated'): False
            ('chat_id'): 'oc_5ad11d72b830411d72b836c20'
            ('sender'): NestedDict(
              ('id'): 'cli_9f427eec54ae901b'
              ('id_type'): 'app_id'
              ('sender_type'): 'app'
              ('tenant_key'): '736588c9260f175e'
            )
            ('body'): NestedDict(
              ('content'): '{"text":"@_user_1 test content"}'
            )
            ('mentions'): [NestedDict(
              ('key'): '@_user_1'
              ('id'): 'ou_155184d1e73cbfb8973e5a9e698e74f2'
              ('id_type'): 'open_id'
              ('name'): 'Chang'
              ('tenant_key'): '736588c9260f175e'
            )]
            ('upper_message_id'): 'om_40eb06e7b84dc71c03e009ad3c754195'
          )
        )
    """  # noqa: E501
    return request(
        "POST",
        dest,
        data=data,
        params=params,
        headers=headers,
        timeout=timeout,
        token=token,
        app_id=app_id,
        app_secret=app_secret,
        **kwargs,
    )


def get(
    dest: str,
    params: Dict | None = None,
    headers: Dict | None = None,
    timeout: int = 120,
    token: str | None = None,
    app_id: str | None = None,
    app_secret: str | None = None,
    **kwargs,
) -> NestedDict:
    r"""
    向指定目的地发送 GET 请求。

    Args:
        dest: 目的地统一资源定位符。这个地址应该是相对于飞书 API 的。
        headers: 请求头。默认为 `{}`。
        params: 请求参数。默认为 `{}`。
        timeout: 超时时间。默认为 120 秒。
        token: 访问凭证。如果为 `None`，则会自动获取。
            如果 token 为 `None`，则 app_id 和 app_secret 必须不为 `None`。
        app_id: 应用唯一标识（以`cli_`开头）。默认为 `None`。
        app_secret: 应用秘钥。默认为 `None`。

    Returns:
        请求结果

    Raises:
        ValueError: 如果 token, app_id, 和 app_secret 都为 `None`。

    Examples:
        >>> get('im/v1/messages/:message_id', app_id='cli_slkdjalasdkjasd', app_secret='dskLLdkasdjlasdKK')  # doctest:+SKIP
        NestedDict(
          ('code'): 0
          ('msg'): 'success'
          ('data'): NestedDict(
            ('items'): [NestedDict(
              ('message_id'): 'om_dc13264520392913993dd051dba21dcf'
              ('root_id'): 'om_40eb06e7b84dc71c03e009ad3c754195'
              ('parent_id'): 'om_d4be107c616aed9c1da8ed8068570a9f'
              ('thread_id'): 'omt_d4be107c616a'
              ('msg_type'): 'card'
              ('create_time'): '1635675360'
              ('update_time'): '1635675360'
              ('deleted'): False
              ('updated'): False
              ('chat_id'): 'oc_5ad11d72b830411d72b836c20'
              ('sender'): NestedDict(
                ('id'): 'cli_9f427eec54ae901b'
                ('id_type'): 'app_id'
                ('sender_type'): 'app'
                ('tenant_key'): '736588c9260f175e'
              )
              ('body'): NestedDict(
                ('content'): '{"text":"@_user_1 test content"}'
              )
              ('mentions'): [NestedDict(
                ('key'): '@_user_1'
                ('id'): 'ou_155184d1e73cbfb8973e5a9e698e74f2'
                ('id_type'): 'open_id'
                ('name'): 'Chang'
                ('tenant_key'): '736588c9260f175e'
              )]
            ('upper_message_id'): 'om_40eb06e7b84dc71c03e009ad3c754195'
            )]
          )
        )
    """  # noqa: E501
    return request(
        "GET",
        dest,
        headers=headers,
        params=params,
        timeout=timeout,
        token=token,
        app_id=app_id,
        app_secret=app_secret,
        **kwargs,
    )


def put(
    dest: str,
    data: Dict | None = None,
    params: Dict | None = None,
    headers: Dict | None = None,
    timeout: int = 120,
    token: str | None = None,
    app_id: str | None = None,
    app_secret: str | None = None,
    **kwargs,
) -> NestedDict:
    r"""
    向指定目的地发送 PUT 请求。

    Args:
        dest: 目的地统一资源定位符。这个地址应该是相对于飞书 API 的。
        headers: 请求头。默认为 `{}`。
        params: 请求参数。默认为 `{}`。
        timeout: 超时时间。默认为 120 秒。
        token: 访问凭证。如果为 `None`，则会自动获取。
            如果 token 为 `None`，则 app_id 和 app_secret 必须不为 `None`。
        app_id: 应用唯一标识（以`cli_`开头）。默认为 `None`。
        app_secret: 应用秘钥。默认为 `None`。

    Returns:
        请求结果

    Raises:
        ValueError: 如果 token, app_id, 和 app_secret 都为 `None`。

    Examples:
        >>> data = {"msg_type": "text", "content": "{\"text\":\"test content\"}"}
        >>> put('im/v1/messages/:message_id', data=data, app_id='cli_slkdjalasdkjasd', app_secret='dskLLdkasdjlasdKK')  # doctest:+SKIP
        NestedDict(
          ('code'): 0
          ('msg'): 'success'
          ('data'): NestedDict(
            ('message_id'): 'om_dc13264520392913993dd051dba21dcf'
            ('root_id'): 'om_40eb06e7b84dc71c03e009ad3c754195'
            ('parent_id'): 'om_d4be107c616aed9c1da8ed8068570a9f'
            ('thread_id'): 'omt_d4be107c616a'
            ('msg_type'): 'card'
            ('create_time'): '1635675360'
            ('update_time'): '1635675420'
            ('deleted'): False
            ('updated'): True
            ('chat_id'): 'oc_5ad11d72b830411d72b836c20'
            ('sender'): NestedDict(
              ('id'): 'cli_9f427eec54ae901b'
              ('id_type'): 'app_id'
              ('sender_type'): 'app'
              ('tenant_key'): '736588c9260f175e'
            )
            ('body'): NestedDict(
              ('content'): '{"text":"@_user_1 test content"}'
            )
            ('mentions'): [NestedDict(
              ('key'): '@_user_1'
              ('id'): 'ou_155184d1e73cbfb8973e5a9e698e74f2'
              ('id_type'): 'open_id'
              ('name'): 'Chang'
              ('tenant_key'): '736588c9260f175e'
            )]
            ('upper_message_id'): 'om_40eb06e7b84dc71c03e00ida3c754892'
          )
        )
    """  # noqa: E501
    return request(
        "PUT",
        dest,
        data=data,
        params=params,
        headers=headers,
        timeout=timeout,
        token=token,
        app_id=app_id,
        app_secret=app_secret,
        **kwargs,
    )


def patch(
    dest: str,
    data: Dict | None = None,
    params: Dict | None = None,
    headers: Dict | None = None,
    timeout: int = 120,
    token: str | None = None,
    app_id: str | None = None,
    app_secret: str | None = None,
    **kwargs,
) -> NestedDict:
    r"""
    向指定目的地发送 PUT 请求。

    Args:
        dest: 目的地统一资源定位符。这个地址应该是相对于飞书 API 的。
        headers: 请求头。默认为 `{}`。
        params: 请求参数。默认为 `{}`。
        timeout: 超时时间。默认为 120 秒。
        token: 访问凭证。如果为 `None`，则会自动获取。
            如果 token 为 `None`，则 app_id 和 app_secret 必须不为 `None`。
        app_id: 应用唯一标识（以`cli_`开头）。默认为 `None`。
        app_secret: 应用秘钥。默认为 `None`。

    Returns:
        请求结果

    Raises:
        ValueError: 如果 token, app_id, 和 app_secret 都为 `None`。

    Examples:
        >>> data = {"content": "{\"text\":\"test content\"}"}
        >>> patch('im/v1/messages/:message_id', data=data, app_id='cli_slkdjalasdkjasd', app_secret='dskLLdkasdjlasdKK')  # doctest:+SKIP
        NestedDict(
          ('code'): 0
          ('msg'): 'success'
          ('data'): NestedDict()
        )
    """  # noqa: E501
    return request(
        "PATCH",
        dest,
        data=data,
        params=params,
        headers=headers,
        timeout=timeout,
        token=token,
        app_id=app_id,
        app_secret=app_secret,
        **kwargs,
    )


def delete(
    dest: str,
    data: Dict | None = None,
    headers: Dict | None = None,
    timeout: int = 120,
    token: str | None = None,
    app_id: str | None = None,
    app_secret: str | None = None,
    **kwargs,
) -> NestedDict:
    r"""
    向指定目的地发送 DELETE 请求。

    Args:
        dest: 目的地统一资源定位符。这个地址应该是相对于飞书 API 的。
        headers: 请求头。默认为 `{}`。
        params: 请求参数。默认为 `{}`。
        timeout: 超时时间。默认为 120 秒。
        token: 访问凭证。如果为 `None`，则会自动获取。
            如果 token 为 `None`，则 app_id 和 app_secret 必须不为 `None`。
        app_id: 应用唯一标识（以`cli_`开头）。默认为 `None`。
        app_secret: 应用秘钥。默认为 `None`。

    Returns:
        请求结果

    Raises:
        ValueError: 如果 token, app_id, 和 app_secret 都为 `None`。

    Examples:
        >>> delete('im/v1/messages/:message_id', app_id='cli_slkdjalasdkjasd', app_secret='dskLLdkasdjlasdKK')  # doctest:+SKIP
        NestedDict(
          ('code'): 0
          ('msg'): 'success'
          ('data'): NestedDict()
        )
    """  # noqa: E501
    return request(
        "DELETE",
        dest,
        headers=headers,
        data=data,
        timeout=timeout,
        token=token,
        app_id=app_id,
        app_secret=app_secret,
        **kwargs,
    )



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# noqa: E302

from __future__ import annotations

from typing import TYPE_CHECKING, Mapping, overload

if TYPE_CHECKING:
    from openai import Stream

@overload
def send_message(
    message: str,
    receive_id: str | None = None,
    receive_id_type: str = "open_id",
    uuid: str = "",
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
        uuid: 消息唯一标识，用于消息去重
    """

@overload
def send_message(
    message: Mapping,
    receive_id: str | None = None,
    receive_id_type: str = "open_id",
    uuid: str = "",
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
        uuid: 消息唯一标识，用于消息去重
    """

@overload
def send_message(
    message: Stream,
    receive_id: str | None = None,
    receive_id_type: str = "open_id",
    uuid: str = "",
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
        uuid: 消息唯一标识，用于消息去重
    """

@overload
def send_message(
    message: str,
    message_id: str,
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
        uuid: 消息唯一标识，用于消息去重
    """

@overload
def send_message(
    message: Mapping,
    message_id: str,
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
        uuid: 消息唯一标识，用于消息去重
    """

@overload
def send_message(
    message: Stream,
    message_id: str,
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
        uuid: 消息唯一标识，用于消息去重
    """



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from functools import wraps
from typing import Callable, List, Mapping, Optional

from . import variables


def authorize(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs):
        # Check if token, app_id, and app_secret are provided in kwargs
        token = kwargs.get("token")
        app_id = kwargs.get("app_id")
        app_secret = kwargs.get("app_secret")

        # If not provided, raise an error or fetch them from somewhere else
        if token is None:
            if app_id is None:
                if not hasattr(variables, "APP_ID"):
                    raise ValueError("app_id is not provided and not found in variables")
                app_id = kwargs["app_id"] = variables.APP_ID
            if app_secret is None:
                if not hasattr(variables, "APP_SECRET"):
                    raise ValueError("app_secret is not provided and not found in variables")
                app_secret = kwargs["app_secret"] = variables.APP_SECRET

        # Call the original function with provided authorization details
        return func(*args, **kwargs)

    return wrapper


def pagination(
    page_size: int = 50, max_pages: Optional[int] = None, condition: Optional[Callable[[List[Mapping]], bool]] = None
) -> Callable:
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(
            env: str,
            space_id: str,
            parent_node_token: str,
            *args,
            **kwargs,
        ) -> Mapping:
            all_items = []
            page_counter = 0
            page_token = None

            while True:
                response = func(
                    env,
                    space_id,
                    parent_node_token,
                    page_size=page_size,
                    page_token=page_token,
                    *args,  # noqa: B026
                    **kwargs,
                )
                items = [i for i in response.data["items"] if i["obj_type"] == "docx"]
                all_items.extend(items)

                page_counter += 1
                if max_pages and page_counter >= max_pages:
                    break

                if condition and condition(all_items):
                    break

                page_token = response.data.get("page_token")
                if not page_token:
                    break

            response.data["items"] = all_items

            return response

        return wrapper

    return decorator



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Mapping

from chanfig import NestedDict

from feishu.api.decorators import authorize, pagination
from feishu.api.request import delete, get, patch, post, put

from .utils import get_stream_message, infer_receive_id_type

if TYPE_CHECKING:
    from openai import Stream


@authorize
def send_message(
    message: str | Mapping | Stream,
    receive_id: str | None = None,
    receive_id_type: str | None = None,
    uuid: str = "",
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

        [回复消息](https://open.feishu.cn/document/server-docs/im-v1/message/reply)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
        uuid: 消息唯一标识，用于消息去重
    """
    if isinstance(message, str):
        message = {
            "msg_type": "text",
            "content": json.dumps({"text": message}),
        }
    if receive_id is not None:
        if receive_id.startswith("om_"):
            return reply_message(message, receive_id, **kwargs)
        message["receive_id"] = receive_id
    receive_id = message.get("receive_id")
    if receive_id is None:
        raise ValueError("receive_id is required")
    if receive_id_type is None:
        receive_id_type = infer_receive_id_type(receive_id)
    if uuid:
        message["uuid"] = uuid
    message = post("im/v1/messages", message, {"receive_id_type": receive_id_type}, **kwargs)
    message.data.body.content = NestedDict.from_jsons(message.data.body.content)
    return message


@authorize
def send_message_stream(
    stream: Stream,
    receive_id: str | None = None,
    receive_id_type: str | None = None,
    uuid: str = "",
    **kwargs,
):
    r"""
    发送消息

    See Also:
        [发送消息](https://open.feishu.cn/document/server-docs/im-v1/message/create)

    Args:
        message: 消息内容
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
    """
    if receive_id_type is None:
        receive_id_type = infer_receive_id_type(receive_id)
    content = ""
    response = send_message(get_stream_message(content, uuid=uuid), **kwargs)
    message_id = response["data"]["message_id"]
    try:
        for chunk in stream:
            content += chunk.choices[0].delta.content
            message = get_stream_message(content, streaming=True, uuid=uuid)
            patch_message(message, message_id, uuid=uuid, **kwargs)
    except Exception as e:
        raise e
    finally:
        message = get_stream_message(content, uuid=uuid)
        patch_message(message, message_id, uuid=uuid, **kwargs)


@authorize
def reply_message(message: str | Mapping, message_id: str, **kwargs):
    r"""
    回复消息

    See Also:
        [回复消息](https://open.feishu.cn/document/server-docs/im-v1/message/reply)

    Args:
        message: 消息内容
        message_id: 消息 ID
    """
    if isinstance(message, str):
        message = {
            "msg_type": "text",
            "content": json.dumps({"text": message}),
        }
    message = post(f"im/v1/messages/{message_id}/reply", message, **kwargs)
    message.data.body.content = NestedDict.from_jsons(message.data.body.content)
    return message


@authorize
def update_message(message: str | Mapping, message_id: str, **kwargs):
    r"""
    编辑消息

    See Also:
        [编辑消息](https://open.feishu.cn/document/server-docs/im-v1/message/update)

    Args:
        message_id: 消息 ID
        message: 消息内容
    """
    if isinstance(message, str):
        message = {
            "msg_type": "text",
            "content": json.dumps({"text": message}),
        }
    message = put(f"im/v1/messages/{message_id}", message, **kwargs)
    message.data.body.content = NestedDict.from_jsons(message.data.body.content)
    return message


@authorize
def recall_message(message_id: str, **kwargs):
    r"""
    撤回消息

    See Also:
        [撤回消息](https://open.feishu.cn/document/server-docs/im-v1/message/recall)

    Args:
        message_id: 消息 ID
    """
    return delete(f"im/v1/messages/{message_id}", **kwargs)


@authorize
def forward_message(message_id: str, receive_id: str, receive_id_type: str | None = None, uuid: str = "", **kwargs):
    r"""
    转发消息

    See Also:
        [转发消息](https://open.feishu.cn/document/server-docs/im-v1/message/forward)

    Args:
        message_id: 消息 ID
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
    """
    if receive_id_type is None:
        receive_id_type = infer_receive_id_type(receive_id)
    message = post(
        f"im/v1/messages/{message_id}/forward",
        {"receive_id": receive_id},
        {"receive_id_type": receive_id_type, "uuid": uuid},
        **kwargs,
    )
    message.data.body.content = NestedDict.from_jsons(message.data.body.content)
    return message


@authorize
def merge_forward(
    message_id_list: list[str],
    receive_id: str,
    receive_id_type: str | None = None,
    uuid: str = "",
    **kwargs,
):
    r"""
    合并转发消息

    See Also:
        [合并转发消息](https://open.feishu.cn/document/server-docs/im-v1/message/merge_forward)

    Args:
        message_id_list: 消息 ID 列表
        receive_id: 接收者 ID
        receive_id_type: 接收者 ID 类型。默认为 `open_id`。
    """
    if receive_id_type is None:
        receive_id_type = infer_receive_id_type(receive_id)
    message = post(
        "im/v1/messages/merge-forward",
        {"message_id_list": message_id_list, "receive_id": receive_id},
        {"receive_id_type": receive_id_type, "uuid": uuid},
        **kwargs,
    )
    message.data.body.content = NestedDict.from_jsons(message.data.body.content)
    return message


@pagination
@authorize
def read_users(message_id: str, user_id_type: str = "open_id", **kwargs):
    r"""
    查询消息已读信息

    See Also:
        [查询消息已读信息](https://open.feishu.cn/document/server-docs/im-v1/message/read_users)

    Args:
        message_id: 消息 ID
        user_id_type: 用户 ID 类型。默认为 `open_id`。可以是 `open_id` 或 "union_id" 或 `user_id`。
    """

    return get(f"im/v1/messages/{message_id}/read_users", {"user_id_type": user_id_type}, **kwargs)


@authorize
def get_message(message_id: str, **kwargs):
    r"""
    获取消息

    See Also:
        [获取指定消息的内容](https://open.feishu.cn/document/server-docs/im-v1/message/get)

    Args:
        message_id: 消息 ID
    """

    message = get(f"im/v1/messages/{message_id}", **kwargs)
    message.data = message.data["items"][0]
    message.data.body.content = NestedDict.from_jsons(message.data.body.content)
    return message


@pagination
@authorize
def get_messages(
    container_id: str,
    container_id_type: str = "chat",
    start_time: int | None = None,
    end_time: int | None = None,
    sort_type: str = "ByCreateTimeAsc",
    page_size: int = 50,
    page_token: str | None = None,
    **kwargs,
):
    r"""
    获取会话历史消息

    See Also:
        [获取会话历史消息](https://open.feishu.cn/document/server-docs/im-v1/message/list)

    Args:
        container_id: 会话 ID
        container_id_type: 会话 ID 类型。默认为 `chat`。可以是 `chat` 或 `group` 或 `meeting`。
        start_time: 开始时间。默认为 `None`。
        end_time: 结束时间。默认为 `None`。
        sort_type: 排序方式。默认为 `ByCreateTimeAsc`。可以是 `ByCreateTimeAsc` 或 `ByCreateTimeDesc`。
        page_size: 每页数量。默认为 50。
        page_token: 分页标识。默认为 `None`。
    """
    message = get(
        "im/v1/messages",
        {
            "container_id": container_id,
            "container_id_type": container_id_type,
            "start_time": start_time,
            "end_time": end_time,
            "sort_type": sort_type,
            "page_size": page_size,
            "page_token": page_token,
        },
        **kwargs,
    )
    message.data["items"] = [NestedDict.from_jsons(i) for i in message.data["items"]]
    return message


@authorize
def get_message_resources(message_id: str, file_key: str, type: str, **kwargs):
    r"""
    获取消息中的资源文件

    See Also:
        [获取消息中的资源文件](https://open.feishu.cn/document/server-docs/im-v1/message/get-2)

    Args:
        message_id: 消息 ID
    """

    return get(f"im/v1/messages/{message_id}/resources/{file_key}?type={type}", **kwargs)


@authorize
def push_follow_up(message_id: str, follow_ups: str | Mapping, **kwargs):
    r"""
    添加跟随气泡

    See Also:
        [添加跟随气泡](https://open.feishu.cn/document/server-docs/im-v1/message/push_follow_up)

    Args:
        message_id: 消息 ID
        follow_ups: 跟随气泡内容
    """

    if isinstance(follow_ups, str):
        follow_ups = {
            "follow_ups": [
                {
                    "content": follow_ups,
                }
            ]
        }
    return post(f"im/v1/messages/{message_id}/push_follow_up", follow_ups, **kwargs)


@authorize
def patch_message(message: str | Mapping, message_id: str, **kwargs):
    r"""
    更新应用发送的消息卡片

    See Also:
        [更新应用发送的消息卡片](https://open.feishu.cn/document/server-docs/im-v1/message-card/patch)

    Args:
        message_id: 消息 ID
        message: 消息内容
    """
    if isinstance(message, str):
        message = {
            "content": json.dumps({"text": message}),
        }
    message = patch(f"im/v1/messages/{message_id}", message, **kwargs)
    message.data.body.content = NestedDict.from_jsons(message.data.body.content)
    return message



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import base64
import hashlib

from Crypto.Cipher import AES


class AESCipher:
    r"""
    AES 加密/解密工具类

    Args:
        key: 用于加密/解密的密钥

    阅读更多:
        [配置 Encrypt Key](https://open.feishu.cn/document/server-docs/event-subscription-guide/event-subscription-configure-/configure-encrypt-key)

    Examples:
        >>> cypher = AESCipher("test key")
        >>> encrypt = "P37w+VZImNgPEO1RBhJ6RtKl7n6zymIbEG1pReEzghk="
        >>> cypher.decrypt(encrypt)
        'hello world'
    """  # noqa: E501

    def __init__(self, key: str):
        self.bs = AES.block_size
        self.key = hashlib.sha256(AESCipher.str_to_bytes(key)).digest()

    @staticmethod
    def str_to_bytes(data: str) -> bytes:
        u_type = type(b"".decode("utf8"))
        if isinstance(data, u_type):
            return data.encode("utf8")
        return data  # type: ignore[return-value]

    @staticmethod
    def _unpad(s: bytes) -> bytes:
        return s[: -ord(s[len(s) - 1 :])]

    def decrypt(self, enc: str | bytes) -> str | bytes:
        if isinstance(enc, str):
            return self._decrypt(base64.b64decode(enc)).decode("utf8")
        if isinstance(enc, bytes):
            return self._decrypt(enc)
        raise ValueError("Invalid type")

    def _decrypt(self, enc: bytes) -> bytes:
        iv = enc[: AES.block_size]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(enc[AES.block_size :]))



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from attrs import define


@define
class FeishuException(Exception):
    r"""
    通用异常

    阅读更多:
        [通用错误码](https://open.feishu.cn/document/server-docs/api-call-guide/generic-error-code)
    """

    code: int
    message: str

    def __repr__(self):
        return f"FeishuException(code={self.code}, message={self.message})"



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import overload

import requests
from chanfig import FlatDict

from feishu.api import variables

from ..exceptions import FeishuException


@overload
def get_app_access_token(app_id: str, app_secret: str, timeout: int = 120) -> FlatDict:
    r"""
    获取应用访问凭证

    Args:
        app_id: 应用唯一标识（以`cli_`开头）
        app_secret: 应用秘钥
        timeout: 超时时间。默认为 120 秒。

    Returns:
        应用访问凭证
    """


@overload
def get_app_access_token(app_id: str, app_secret: str, app_ticket: str, timeout: int = 120) -> FlatDict:
    r"""
    获取应用访问凭证

    Args:
        app_id: 应用唯一标识（以`cli_`开头）
        app_secret: 应用秘钥
        app_ticket: 平台定时推送给应用的临时凭证
        timeout: 超时时间。默认为 120 秒。

    Returns:
        应用访问凭证
    """


def get_app_access_token(  # type: ignore[misc]
    app_id: str,
    app_secret: str,
    app_ticket: str | None = None,
    timeout: int = 120,
) -> FlatDict:
    r"""
    获取应用访问凭证

    这是一个通用函数，根据传入的参数自动判断是自建应用还是商店应用，并调用相应的函数。

    Args:
        app_id: 应用唯一标识（以`cli_`开头）
        app_secret: 应用秘钥
        app_ticket: 平台定时推送给应用的临时凭证
        timeout: 超时时间。默认为 120 秒。

    Returns:
        应用访问凭证

    阅读更多:
        [通用参数介绍](https://open.feishu.cn/document/server-docs/api-call-guide/terminology)

        获取自建应用的应用访问凭证：[feishu.api.get_app_access_token_internal][]

        获取商店应用的应用访问凭证：[feishu.api.get_app_access_token_store][]

    Examples:
        >>> get_app_access_token("cli_slkdjalasdkjasd", "dskLLdkasdjlasdKK")  # doctest:+SKIP
        FlatDict(
          ('app_access_token'): 't-g1044ghJRUIJJ5ZPPZMOHKWZISL33E4QSS3abcef'
          ('code'): 0
          ('expire'): 7200
          ('msg'): 'ok'
          ('tenant_access_token'): 't-g1044ghJRUIJJ5ZPPZMOHKWZISL33E4QSS3abcef'
        )
        >>> get_app_access_token("cli_slkdjalasdkjasd", "dskLLdkasdjlasdKK", "dskLLdkasd")  # doctest:+SKIP
        FlatDict(
          ('code'): 0
          ('msg'): 'success'
          ('app_access_token'): 'a-6U1SbDiM6XIH2DcTCPyeub'
          ('expire'): 7200
        )
    """
    if app_ticket is None:
        return get_app_access_token_internal(app_id, app_secret, timeout)
    if isinstance(app_ticket, int):
        return get_app_access_token_internal(app_id, app_secret, app_ticket)
    return get_app_access_token_store(app_id, app_secret, app_ticket, timeout)


def get_app_access_token_internal(app_id: str, app_secret: str, timeout: int = 120) -> FlatDict:
    r"""
    自建应用获取应用访问凭证

    Args:
        app_id: 应用唯一标识（以`cli_`开头）
        app_secret: 应用秘钥
        timeout: 超时时间。默认为 120 秒。

    Returns:
        应用访问凭证

    阅读更多:
        [通用参数介绍](https://open.feishu.cn/document/server-docs/api-call-guide/terminology)

        [自建应用获取应用访问凭证](https://open.feishu.cn/document/server-docs/authentication-management/access-token/app_access_token_internal)

    Examples:
        >>> get_app_access_token_internal("cli_slkdjalasdkjasd", "dskLLdkasdjlasdKK")  # doctest:+SKIP
        FlatDict(
          ('app_access_token'): 't-g1044ghJRUIJJ5ZPPZMOHKWZISL33E4QSS3abcef'
          ('code'): 0
          ('expire'): 7200
          ('msg'): 'ok'
          ('tenant_access_token'): 't-g1044ghJRUIJJ5ZPPZMOHKWZISL33E4QSS3abcef'
        )
    """
    url = variables.BASE_URL + "auth/v3/app_access_token/internal"
    data = {"app_id": app_id, "app_secret": app_secret}
    response = requests.post(url, json=data, timeout=timeout)
    if response.status_code != 200:
        raise FeishuException(response.status_code, response.text)
    ret = response.json()
    if ret.get("code") != 0:
        raise FeishuException(ret.get("code"), ret.get("msg"))
    return FlatDict(ret)


def get_app_access_token_store(app_id: str, app_secret: str, app_ticket: str, timeout: int = 120) -> FlatDict:
    r"""
    商店应用获取应用访问凭证

    Args:
        app_id: 应用唯一标识（以`cli_`开头）
        app_secret: 应用秘钥
        app_ticket: 平台定时推送给应用的临时凭证
        timeout: 超时时间。默认为 120 秒。

    Returns:
        应用访问凭证

    阅读更多:
        [通用参数介绍](https://open.feishu.cn/document/server-docs/api-call-guide/terminology)

        [商店应用获取应用访问凭证](https://open.feishu.cn/document/server-docs/authentication-management/access-token/app_access_token)

    Examples:
        >>> get_app_access_token_store("cli_slkdjalasdkjasd", "dskLLdkasdjlasdKK", "dskLLdkasd")  # doctest:+SKIP
        FlatDict(
          ('code'): 0
          ('msg'): 'success'
          ('app_access_token'): 'a-6U1SbDiM6XIH2DcTCPyeub'
          ('expire'): 7200
        )
    """  # noqa: E501
    url = variables.BASE_URL + "auth/v3/app_access_token"
    data = {"app_id": app_id, "app_secret": app_secret, "app_ticket": app_ticket}
    response = requests.post(url, json=data, timeout=timeout)
    if response.status_code != 200:
        raise FeishuException(response.status_code, response.text)
    ret = response.json()
    if ret.get("code") != 0:
        raise FeishuException(ret.get("code"), ret.get("msg"))
    return FlatDict(ret)



# open-feishu
# Copyright (C) 2024-Present  Zhiyuan Chen

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from email.utils import parseaddr
from json import dumps

from feishu.api import variables


def infer_receive_id_type(receive_id: str) -> str:
    r"""
    根据 receive_id 来推断 receive_id_type
    """
    if receive_id.startswith("ou_"):
        return "open_id"
    if receive_id.startswith("on_"):
        return "union_id"
    if receive_id.startswith("oc_"):
        return "chat_id"
    if len(receive_id) == 8:
        return "user_id"
    if parseaddr(receive_id)[1]:
        return "email"
    raise ValueError("unable to infer receive_id_type")


def get_stream_message(content: str, streaming: bool = False, uuid: str = ""):
    elements = [{"tag": "div", "text": {"content": content, "tag": "lark_md"}}]

    if streaming:
        elements.append(
            {
                "tag": "div",
                "text": {
                    "content": variables.STREAMING_STATUS_TEXT,
                    "tag": "plain_text",
                    "text_size": "notation",
                    "text_color": "gray",
                    "text_align": "right",
                },
            }
        )

    return {
        "content": dumps({"config": {"wide_screen_mode": True}, "elements": elements}),
        "msg_type": "interactive",
        "uuid": uuid,
    }
